#include <stdio.h>
#include <cstdarg>
#include <iostream>

bool print_debug = true;
FILE* g_log_file = 0;

void debug( const char* format, ... ) {
	if(!print_debug)
		return;
	printf("DEBUG: ");
	va_list args;
	va_start( args, format );
	vprintf(format, args);
	if (g_log_file){
		fprintf(g_log_file, "DEBUG: ");
		vfprintf(g_log_file, format, args);
		fprintf(g_log_file, "\n");
		fflush(g_log_file);
	}
	va_end( args );
	printf("\n");
	fflush(stdout);
}

#if defined _WIN32 || defined __CYGWIN32__

#if defined(_MSC_VER)
#	define for if (false) {} else for
#endif

#include <windows.h>
//#include <conio.h>

// this always goes directly to the blocking read, whether the first 'resetting' wait is there or not
bool sleep_and_input(char* c, int sleep)
{
	void *input = GetStdHandle(STD_INPUT_HANDLE);
	debug("resetting handle");
	WaitForSingleObject(input, 0);
	debug("starting to wait");
	int wait_result = WaitForSingleObject(input, sleep*1000);
	debug("finished waiting");
	if (wait_result ==  WAIT_TIMEOUT) {
		// it timed out - no input
		debug("wait timed out, no input");
		return false;
	} else if (wait_result == WAIT_OBJECT_0){
		debug("recieved input, calling blocking read");
		// never returns when not running from console
		// *c = _getch();
		// waits for enter to be pressed, includes newline in output
		// std::cin.get(*c);
		// *c = std::cin.get();
		// *c = getc(stdin);
		// waits for enter to be pressed, doesn't include newline
		std::cin >> c;
		debug("blocking read returned '%c', returning true", *c);
		return true;
	} else if (wait_result == WAIT_FAILED || wait_result == WAIT_ABANDONED) {
		// some shit went wrong
		debug("somethting went wrong with wait");
		return false;
	}
	return false;
};
/*

bool sleep_and_input(char* c, int sleep)
{
	debug("in sleep_and_input");
	for (int i = 0; i < sleep * 2; ++i)
	{
		// this always just returns 0
		int size = std::cin.rdbuf()->in_avail();
		debug("cin size: '%d'", size);
		if (size > 0)
		{
			debug("reading character now");
			// *c = fgetc(stdin);
			// *c = getchar();

			// waits for enter to be pressed, doesn't include newline
			std::cin >> c;
			debug("character returned '%c', returning true", *c);
			return true;
		}
		Sleep(500);
	}
	debug("returning false");
	return false;
};

// this works when ran from the windows console, but not any other way
// such as if the process is spawned from java, or ran over cygwin ssh
bool sleep_and_input(char* c, int sleep)
{
	for (int i = 0; i < sleep * 2; ++i)
	{
		if (_kbhit())
		{
			*c = _getch();
			return true;
		}
		Sleep(500);
	}
	return false;
};

// this always says "PeekConsoleInput failed.", presumably from the same reason _kbhit never returns true
// it must be looking in a different buffer than stdin
void peek_print_keypress()
{
  HANDLE input_handle = GetStdHandle(STD_INPUT_HANDLE);
  DWORD events = 0;      // how many events took place
  INPUT_RECORD input_record;  // a record of input events
  DWORD input_size = 1;    // how many characters to read

  BOOL peeked = PeekConsoleInput(input_handle, &input_record, input_size, &events);
  if(!peeked)
  {
    debug("PeekConsoleInput failed.");
    return;
  }

  if(peeked && input_record.EventType == KEY_EVENT && input_record.Event.KeyEvent.bKeyDown)
  { // PeekConsoleInput succeeded and a key was pressed, so set control_key and return vk.
	 debug("control key state: '%d', keycode: '%d'", input_record.Event.KeyEvent.dwControlKeyState, input_record.Event.KeyEvent.wVirtualKeyCode);
	 return;
  }

  debug("Key wasn't pressed.");
}
*/
#else

#include <stdlib.h>
#include <cerrno>

#include <termios.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/select.h>

struct set_keypress
{
	set_keypress()
	{
		termios new_settings;
		tcgetattr(0,&stored_settings);
		new_settings = stored_settings;
		// Disable canonical mode, and set buffer size to 1 byte
		new_settings.c_lflag &= (~ICANON);
		new_settings.c_cc[VTIME] = 0;
		new_settings.c_cc[VMIN] = 1;
		tcsetattr(0,TCSANOW,&new_settings);
	}
	~set_keypress() { tcsetattr(0,TCSANOW,&stored_settings); }
	termios stored_settings;
};

// this code works perfectly under Linux and Mac OSX
bool sleep_and_input(char* c, int sleep)
{
	// sets the terminal to single-character mode
	// and resets when destructed
	set_keypress s;
	int ret = 0;

	fd_set set;
	FD_ZERO(&set);
	FD_SET(0, &set);
	timeval tv = {sleep, 0};
	ret = select(1, &set, 0, 0, &tv);
	if (ret > 0)
	{
		*c = getc(stdin);
		return true;
	}
	if (errno == EINTR)
	{
		debug("select failed: %s\n", strerror(errno));
		return false;
	}

	if (ret < 0 && errno != 0 && errno != ETIMEDOUT)
		debug("select failed: %s\n", strerror(errno));

	return false;
}

#endif

int main()
{
	int refresh_delay = 1;
	// main loop
	while (true)
	{

		char c = 0;
		while (sleep_and_input(&c, refresh_delay))
		{
			//if(c == '\n') break;
			debug("character read: '%c'", c);

			if (c == 'q'){
				break;
			}

			debug("another key was pressed, handling it...");
		}

		if (c == 'q') break;

		debug("processing stuff... (Press 'q' to quit.)");
		debug("--------------------------------------------------------");
	}

    return 0;
}
